@@ -105,7 +105,8 @@ match_theme_image (GtkStyle       *style,
 }
 
 static GdkBitmap *
-get_window_for_shape (ThemeImage *image, GdkWindow *window, GtkWidget *widget)
+get_window_for_shape (ThemeImage *image, GdkWindow *window, GtkWidget *widget,
+		      gint x, gint y, gint width, gint height)
 {
   /* It's not a good idea to set a shape mask when painting on anything but
    * widget->window, not only does the shape mask get wrongly offset but also
@@ -117,6 +118,22 @@ get_window_for_shape (ThemeImage *image, GdkWindow *window, GtkWidget *widget)
    */
   if (image->background_shaped && window == widget->window)
     {
+      gint window_width;
+      gint window_height;
+      /* It also is not a good idea to apply shape mask when filling a window
+       * partially.
+       *
+       * For example GtkMenu paints/painted the arrow backgrounds with exactly
+       * the same parameters as full window background - except for the
+       * geometry. (http://bugzilla.gnome.org/show_bug.cgi?id=404571)
+       */
+      if (x != 0 || y != 0)
+        return NULL;
+
+      gdk_drawable_get_size (window, &window_width, &window_height);
+      if (width != window_width || height != window_height)
+        return NULL;
+
       if (GTK_IS_MENU (widget))
 	return gtk_widget_get_parent_window (widget);
       else if (GTK_IS_WINDOW (widget))
@@ -233,7 +250,7 @@ draw_simple_image(GtkStyle       *style,
 	  GdkBitmap *mask = NULL;
 	  gboolean valid;
 
-	  maskwin = get_window_for_shape (image, window, widget);
+	  maskwin = get_window_for_shape (image, window, widget, x, y, width, height);
 	  if (maskwin)
 	    mask = gdk_pixmap_new (maskwin, width, height, 1);
 
@@ -990,6 +1007,55 @@ draw_box_gap (GtkStyle       *style,
 				x, y, width, height, gap_side, gap_x, gap_width);
 }
 
+static void
+draw_expander (GtkStyle        *style,
+               GdkWindow       *window,
+               GtkStateType     state,
+               GdkRectangle    *area,
+               GtkWidget       *widget,
+               const gchar     *detail,
+               gint             center_x,
+               gint             center_y,
+               GtkExpanderStyle expander_style)
+{
+  ThemeMatchData match_data;
+  gint expander_size;
+
+  g_return_if_fail(style != NULL);
+  g_return_if_fail(window != NULL);
+
+  /* Reusing the arrow theming here as it's flexible enough (almost, we do lose
+   * the intermediate states.) It also allows us to use existing gtkrc.
+   * XXX Might want to introduce proper keywords for expanders some day.
+   */
+
+  gtk_widget_style_get (widget, "expander-size", &expander_size, NULL);
+
+  match_data.function = TOKEN_D_ARROW;
+  match_data.detail = (gchar *)detail;
+  match_data.flags = THEME_MATCH_STATE | THEME_MATCH_ARROW_DIRECTION;
+  match_data.state = state;
+
+  switch (expander_style)
+    {
+    case GTK_EXPANDER_COLLAPSED:
+    case GTK_EXPANDER_SEMI_COLLAPSED:
+      match_data.arrow_direction = GTK_ARROW_RIGHT;
+      break;
+    case GTK_EXPANDER_EXPANDED:
+    case GTK_EXPANDER_SEMI_EXPANDED:
+      match_data.arrow_direction = GTK_ARROW_DOWN;
+      break;
+    default:
+      g_return_if_reached ();
+    }
+
+  if (!draw_simple_image (style, window, area, widget, &match_data, TRUE,
+                          center_x - expander_size/2, center_y - expander_size/2, expander_size, expander_size))
+    parent_class->draw_expander (style, window, state, area, widget, detail,
+                                 center_x, center_y, expander_style);
+}
+
 static void
 draw_extension (GtkStyle       *style,
 		GdkWindow      *window,
@@ -1242,4 +1308,5 @@ sapwood_style_class_init (SapwoodStyleClass *klass)
   style_class->draw_slider = draw_slider;
   style_class->draw_handle = draw_handle;
   style_class->render_icon = render_icon;
+  style_class->draw_expander = draw_expander;
 }
